import PropTypes from 'prop-types';
import React, { createContext, useRef, useCallback, useMemo, useContext, useReducer, useEffect, useLayoutEffect } from 'react';

const DataLoaderContext = createContext();

const DataLoaderProvider = ({
  children
}) => {
  const cachedData = useRef({});
  const reloads = useRef({});
  const setCachedData = useCallback(compute => {
    if (typeof compute === 'function') {
      cachedData.current = compute(cachedData.current);
    } else {
      cachedData.current = compute;
    }
  }, []);
  const setReloads = useCallback(compute => {
    if (typeof compute === 'function') {
      reloads.current = compute(reloads.current);
    } else {
      reloads.current = compute;
    }
  }, []);
  const addCachedData = useCallback((key, newData) => {
    if (key && typeof key === 'string' && newData) {
      setCachedData(actualCachedData => ({ ...actualCachedData,
        [key]: newData
      }));
    }
  }, [setCachedData]);
  const addReload = useCallback((key, method) => {
    if (key && typeof key === 'string' && method) {
      setReloads(actualReloads => ({ ...actualReloads,
        [key]: method
      }));
    }
  }, [setReloads]);
  const clearReload = useCallback(key => {
    if (key && typeof key === 'string') {
      setReloads(actualReloads => {
        const tmp = actualReloads;
        delete tmp[key];
        return tmp;
      });
    }
  }, [setReloads]);
  const clearAllReloads = useCallback(() => {
    setReloads({});
  }, [setReloads]);
  const clearCachedData = useCallback(key => {
    if (key && typeof key === 'string') {
      setCachedData(actualCachedData => {
        const tmp = actualCachedData;
        delete tmp[key];
        return tmp;
      });
    }
  }, [setCachedData]);
  const clearAllCachedData = useCallback(() => {
    setCachedData({});
  }, [setCachedData]);
  const reload = useCallback(async key => {
    if (key && typeof key === 'string') {
      await (reloads.current[key] && reloads.current[key]());
    }
  }, []);
  const reloadAll = useCallback(async () => {
    await Promise.all(Object.values(reloads.current).map(reloadFn => reloadFn()));
  }, []);
  const getCachedData = useCallback(key => {
    if (key) {
      return cachedData.current[key] || undefined;
    }

    return cachedData.current;
  }, []);
  const getReloads = useCallback(key => {
    if (key) {
      return reloads.current[key] || undefined;
    }

    return reloads.current;
  }, []);
  const value = useMemo(() => ({
    addCachedData,
    addReload,
    clearAllCachedData,
    clearAllReloads,
    clearCachedData,
    clearReload,
    getCachedData,
    getReloads,
    reload,
    reloadAll
  }), [addCachedData, clearReload, clearCachedData, getCachedData, getReloads, addReload, reload, clearAllCachedData, clearAllReloads, reloadAll]);
  return /*#__PURE__*/React.createElement(DataLoaderContext.Provider, {
    value: value
  }, children);
};

DataLoaderProvider.propTypes = {
  children: PropTypes.node.isRequired
};
const useDataLoaderContext = () => useContext(DataLoaderContext);

const StatusEnum = {
  ERROR: 'error',
  IDLE: 'idle',
  LOADING: 'loading',
  SUCCESS: 'success'
};
const ActionEnum = {
  ON_ERROR: 'ON_ERROR',
  ON_LOADING: 'ON_LOADING',
  ON_SUCCESS: 'ON_SUCCESS',
  ON_UPDATE_DATA: 'ON_UPDATE_DATA',
  RESET: 'RESET'
};

/* eslint-disable consistent-return */
var reducer = ((state, action) => {
  switch (action.type) {
    case ActionEnum.ON_LOADING:
      return { ...state,
        error: undefined,
        status: StatusEnum.LOADING
      };

    case ActionEnum.ON_SUCCESS:
      return { ...state,
        error: undefined,
        status: StatusEnum.SUCCESS
      };

    case ActionEnum.RESET:
      return {
        error: undefined,
        status: StatusEnum.IDLE
      };

    case ActionEnum.ON_ERROR:
      return { ...state,
        error: action.error,
        status: StatusEnum.ERROR
      };
  }
});

const Actions = {
  createOnError: error => ({
    error,
    type: ActionEnum.ON_ERROR
  }),
  createOnLoading: () => ({
    type: ActionEnum.ON_LOADING
  }),
  createOnSuccess: () => ({
    type: ActionEnum.ON_SUCCESS
  }),
  createReset: () => ({
    type: ActionEnum.RESET
  })
};
/**
 * @typedef {Object} useDataLoaderConfig
 * @property {Function} [onSuccess] callback when a request success
 * @property {Function} [onError] callback when a error is occured
 * @property {*} [initialData] initial data if no one is present in the cache before the request
 * @property {number} [pollingInterval] relaunch the request after the last success
 * @property {boolean} [enabled=true] launch request automatically (default true)
 * @property {boolean} [keepPreviousData=true] do we need to keep the previous data after reload (default true)
 */

/**
 * @typedef {Object} useDataLoaderResult
 * @property {boolean} isIdle true if the hook in initial state
 * @property {boolean} isLoading true if the request is launched
 * @property {boolean} isSuccess true if the request success
 * @property {boolean} isError true if the request throw an error
 * @property {boolean} isPolling true if the request if enabled is true, pollingInterval is defined and the status is isLoading or isSuccess
 * @property {*} previousData if keepPreviousData is true it return the last data fetched
 * @property {*} data initialData if no data is fetched or not present in the cache otherwise return the data fetched
 * @property {string} error the error occured during the request
 * @property {Function} reload reload the data
 */

/**
 * @param {string} key key to save the data fetched in a local cache
 * @param {() => Promise} method a method that return a promise
 * @param {useDataLoaderConfig} config hook configuration
 * @returns {useDataLoaderResult} hook result containing data, request state, and method to reload the data
 */

const useDataLoader = (key, method, {
  enabled = true,
  initialData,
  keepPreviousData = true,
  onError,
  onSuccess,
  pollingInterval
} = {}) => {
  const {
    addReload,
    clearReload,
    getCachedData,
    addCachedData
  } = useDataLoaderContext();
  const [{
    status,
    error
  }, dispatch] = useReducer(reducer, {
    error: undefined,
    status: StatusEnum.IDLE
  });
  const addReloadRef = useRef(addReload);
  const clearReloadRef = useRef(clearReload);
  const previousDataRef = useRef();
  const isMountedRef = useRef(enabled);
  const methodRef = useRef(method);
  const onSuccessRef = useRef(onSuccess);
  const onErrorRef = useRef(onError);
  const isLoading = useMemo(() => status === StatusEnum.LOADING, [status]);
  const isIdle = useMemo(() => status === StatusEnum.IDLE, [status]);
  const isSuccess = useMemo(() => status === StatusEnum.SUCCESS, [status]);
  const isError = useMemo(() => status === StatusEnum.ERROR, [status]);
  const isPolling = useMemo(() => enabled && pollingInterval && (isSuccess || isLoading), [isSuccess, isLoading, enabled, pollingInterval]);
  const handleRequest = useRef(async cacheKey => {
    try {
      var _methodRef$current, _onSuccessRef$current;

      dispatch(Actions.createOnLoading());
      const result = await ((_methodRef$current = methodRef.current) === null || _methodRef$current === void 0 ? void 0 : _methodRef$current.call(methodRef));

      if (keepPreviousData) {
        previousDataRef.current = getCachedData(cacheKey);
      }

      if (result !== undefined && result !== null && cacheKey) addCachedData(cacheKey, result);
      dispatch(Actions.createOnSuccess());
      await ((_onSuccessRef$current = onSuccessRef.current) === null || _onSuccessRef$current === void 0 ? void 0 : _onSuccessRef$current.call(onSuccessRef, result));
    } catch (err) {
      var _onErrorRef$current;

      dispatch(Actions.createOnError(err));
      await ((_onErrorRef$current = onErrorRef.current) === null || _onErrorRef$current === void 0 ? void 0 : _onErrorRef$current.call(onErrorRef, err));
    }
  });
  useEffect(() => {
    let handler;

    if (isMountedRef.current) {
      if (isIdle) {
        handleRequest.current(key);
      }

      if (pollingInterval && isSuccess) {
        handler = setTimeout(() => handleRequest.current(key), pollingInterval);
      }
    }

    return () => {
      if (handler) clearTimeout(handler);
    }; // Why can't put empty array for componentDidMount, componentWillUnmount ??? No array act like componentDidMount and componentDidUpdate
  }, [key, pollingInterval, isIdle, isSuccess]);
  useLayoutEffect(() => {
    methodRef.current = method;
  }, [method]);
  useLayoutEffect(() => {
    isMountedRef.current = enabled;
    dispatch(Actions.createReset());

    if (key && typeof key === 'string') {
      var _addReloadRef$current;

      (_addReloadRef$current = addReloadRef.current) === null || _addReloadRef$current === void 0 ? void 0 : _addReloadRef$current.call(addReloadRef, key, reloadArgs => handleRequest.current(key, reloadArgs));
    }

    return () => {
      if (key && typeof key === 'string') {
        var _clearReloadRef$curre;

        (_clearReloadRef$curre = clearReloadRef.current) === null || _clearReloadRef$curre === void 0 ? void 0 : _clearReloadRef$curre.call(clearReloadRef, key);
      }
    };
  }, [enabled, key]);
  useLayoutEffect(() => {
    clearReloadRef.current = clearReload;
    addReloadRef.current = addReload;
  }, [clearReload, addReload]);
  return {
    data: getCachedData(key) || initialData,
    error,
    isError,
    isIdle,
    isLoading,
    isPolling,
    isSuccess,
    previousData: previousDataRef.current,
    reload: args => handleRequest.current(key, args)
  };
};

export { DataLoaderProvider, useDataLoader, useDataLoaderContext };
